# Decorator for automatic pre-start and post-stop
import functools
from functools import wraps

from loguru import logger as log


def spark_session_management(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self.pre_start()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.post_stop()

    return wrapper


# Decorator for logging and exception handling
def log_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            log.debug(f"Executing {func.__name__ if hasattr(func, '__name__') else 'property'}")
            return func(*args, **kwargs)
        except Exception as e:
            log.error(f"Exception in {func.__name__ if hasattr(func, '__name__') else 'property'}: {e}")
            raise

    return wrapper


def connection_check(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check if the connection to ClickHouse is established
        if not self._check_connection():
            raise Exception("ClickHouse connection not established.")
        return func(self, *args, **kwargs)

    return wrapper



from typing import Any, Dict, List, Optional

from pyspark.sql import SparkSession

from seedspark.sparkapp import SparkApps


class SimpleLineCountApp(SparkApps):
    def __init__(  # noqa: PLR0913
        self,
        app_name: str,
        file_path: str,
        extra_packages: Optional[List[str]] = None,
        extra_jars: Optional[List[str]] = None,
        extra_configs: Optional[Dict[str, Any]] = None,
        spark_master: Optional[str] = None,
        environment="staging",
        log_env: bool = True,
        existing_spark_session: Optional[SparkSession] = None,  # New parameter to accept an existing session
    ):
        self.file_path = file_path
        super().__init__(
            app_name,
            extra_packages=extra_packages,
            extra_jars=extra_jars,
            extra_configs=extra_configs,
            spark_master=spark_master,
            environment=environment,
            log_env=log_env,
            existing_spark_session=existing_spark_session,
        )

    def execute(self):
        # Load data
        data = self.spark.read.text(self.file_path)
        # Count the number of lines
        line_count = data.count()
        print(f"Number of lines in the file: {line_count}")
        return line_count


# line_count_app = SimpleLineCountApp(app_name="a",
#                                     file_path = ".")
# line_count_app.execute()



import os
from pathlib import Path
from typing import Optional

import sqlglot
from loguru import logger as log

from seedspark.apps import SparkDeltaClickhouseApp


def parse_sql_file(sql_file_path: str, read_dialect: str = "clickhouse", write_dialect: str = "clickhouse") -> str:
    """Parses an SQL file and returns the optimized SQL query.

    Args:
        sql_file_path: The path to the SQL file.

    Returns:
        The optimized SQL query as a string.

    Raises:
        FileNotFoundError: If the SQL file is not found.
        sqlglot.errors.ParseError: If there is an error parsing the SQL file.
    """

    if not os.path.exists(sql_file_path):
        raise FileNotFoundError(f"SQL file not found: {sql_file_path}")

    with open(sql_file_path) as f:
        query = f.read()

    try:
        sql = sqlglot.transpile(query, read=read_dialect, write=write_dialect, identify=True, pretty=True)[0]
        return sql
    except sqlglot.errors.ParseError as e:
        raise ValueError(f"Error parsing SQL file: {e.errors}") from e


class WeekendFaresKPIApp:
    """
    Class to compute weekend metrics using SparkDeltaClickhouseApp.
    """

    def __init__(
        self,
        app_name="WeekendFaresKPIApp",
        environment="prod",
        sql_file_path: Optional[str] = None,
        sqllite_db_path: Optional[str] = None,
    ):
        self.clickhouse_app = SparkDeltaClickhouseApp(
            app_name=app_name, environment=environment, extra_packages=["org.xerial:sqlite-jdbc:3.45.1.0"]
        )
        root_dir_path = Path(__file__).parent.absolute().__str__()
        if sql_file_path is None:
            sql_file_path = f"{root_dir_path}/sql/weekend_trip_metrics.sql"
        self.sql_query = parse_sql_file(sql_file_path)

        # JDBC URL for SQLite
        if sqllite_db_path is None:
            sqllite_db_path = "weekend_fares_kpi.db"
        self.sqlite_jdbc_url = f"jdbc:sqlite:{sqllite_db_path}"
        self.table_name = "weekend_fares_kpi"

    def execute(self):
        """
        Execute specific metrics calculations for weekends and return DataFrame.
        """
        # Example SQL query
        df = self.clickhouse_app.execute(self.sql_query)

        # Write to SQLite using JDBC
        log.info(f"Writing to SQLite table: {self.table_name} with JDBC URL: {self.sqlite_jdbc_url}")
        df.write.format("jdbc").option("url", self.sqlite_jdbc_url).option("dbtable", self.table_name).option(
            "driver", "org.sqlite.JDBC"
        ).mode("overwrite").save()


if __name__ == "__main__":
    weekendApp = WeekendFaresKPIApp()
    weekendApp.execute()



import importlib.metadata
import os
import subprocess
import urllib
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import requests
from loguru import logger as log
from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession

from seedspark.decorators import log_exceptions

# Constants
HTTP_STATUS_OK = 200
REQUEST_TIMEOUT = 10  # in seconds
HTTP_SCHEMES = ["http", "https"]
JAR_DIR_DEFAULT = "jars"
DELTA_MAVEN_FORMAT = "io.delta:delta-spark_{scala_version}:{delta_version}"


class SparkAppsException(Exception):
    """Custom Exception for SparkApps."""


class SparkApps(ABC):
    """Base class for Spark applications with optimized configurations and functionalities."""

    def __init__(  # noqa: PLR0913
        self,
        app_name: str,
        extra_packages: Optional[List[str]] = None,
        extra_jars: Optional[List[str]] = None,
        extra_configs: Optional[Dict[str, Any]] = None,
        spark_master: Optional[str] = None,
        environment="staging",
        log_env: bool = True,
        scala_version: str = "2.12",
        enable_delta_jar: bool = False,
        delta_version: str = "3.1.0",
        jar_dir: str = JAR_DIR_DEFAULT,
        existing_spark_session: Optional[SparkSession] = None,  # New parameter to accept an existing session
    ) -> None:
        self.app_name = app_name
        if enable_delta_jar:
            self.extra_configs = {**(self._apply_delta_config()), **(extra_configs or {})}
        else:
            self.extra_configs = extra_configs or {}
        self.spark_master = spark_master
        self._sc = None
        self.jar_dir = jar_dir
        self.delta_version = delta_version
        self.scala_version = scala_version
        self.enable_delta_jar = enable_delta_jar
        self.environment = environment
        self.extra_jars = self.resolve_jar_urls(extra_jars) if extra_jars else []
        self.extra_packages = extra_packages or []
        if log_env:
            self.log_sys_env()
        self._spark = existing_spark_session
        self._spark = self._initialize_spark_session()  # Use existing session if provided

    @property
    @log_exceptions
    def spark_conf(self) -> SparkConf:
        """Property to get SparkConf object with all configurations set."""
        spark_conf = SparkSession.builder.appName(self.app_name)
        if self.spark_master:
            spark_conf.master(self.spark_master)
        if self.extra_packages or self.enable_delta_jar:
            packages = self._build_maven_packages()
            spark_conf.config("spark.jars.packages", packages)
            log.info(f"Setting spark.jars.packages to {packages}")
        if self.extra_jars:
            spark_conf.config("spark.jars", ",".join(self.resolve_jar_urls(self.extra_jars)))
        for key, value in self.extra_configs.items():
            spark_conf.config(key, value)
        self._set_adaptive_configs(spark_conf)
        return spark_conf

    @abstractmethod
    def execute(self):
        """Abstract method to be implemented by subclasses."""
        raise NotImplementedError("execute method must be implemented")

    @property
    @log_exceptions
    def spark(self) -> SparkSession:
        """Lazily initializes and returns the SparkSession object."""
        if not self._spark:
            self._spark = self._initialize_spark_session()
        return self._spark

    @property
    @log_exceptions
    def sc(self) -> SparkContext:
        """Lazily initializes and returns the SparkContext object."""
        return self.spark.sparkContext

    @log_exceptions
    def _initialize_spark_session(self) -> SparkSession:
        """Lazily initializes and returns the SparkSession object."""
        if not self._spark:
            _spark_conf = self.spark_conf  # Corrected to use it as a property, not a method
            log.info(f"Initializing SparkSession with SparkConf: {_spark_conf}")
            _spark = _spark_conf.getOrCreate()
            log.info(
                f"SparkSession initialized with version: {_spark.version} and configurations: {_spark.sparkContext.getConf().getAll()}"
            )
            return _spark
        return self._spark

    @log_exceptions
    def _set_adaptive_configs(self, spark_conf: SparkConf) -> None:
        """Sets adaptive query execution configurations if enabled."""
        if self.extra_configs.get("enableAdaptiveQueryExecution", False):
            spark_conf.config("spark.sql.adaptive.enabled", "true")
            spark_conf.config("spark.sql.adaptive.coalescePartitions.enabled", "true")
            spark_conf.config("spark.sql.adaptive.skewJoin.enabled", "true")

    @log_exceptions
    def stop_spark_session(self) -> None:
        """Stops the Spark session if initialized."""
        if self._spark:
            self._spark.stop()

    @staticmethod
    @log_exceptions
    def log_sys_env() -> None:
        """Logs the Java version and installed Python packages."""
        try:
            java_version = subprocess.check_output(["java", "-version"], stderr=subprocess.STDOUT).decode()
            log.info(f"Java Version: {java_version}")
        except subprocess.CalledProcessError as e:
            log.error(f"Error obtaining Java version: {e}")
        installed_packages_list = sorted(
            f"{dist.metadata['Name']}=={dist.version}" for dist in importlib.metadata.distributions()
        )
        log.info(f"Installed packages: {installed_packages_list}")

    @log_exceptions
    def resolve_jar_urls(self, jar_urls) -> List[str]:
        """Resolves JAR URLs, downloading if necessary."""
        if not isinstance(jar_urls, list) or any(not isinstance(url, str) for url in jar_urls):
            raise TypeError("jar_urls must be a list of strings.")

        resolved_jars = []
        for jar in jar_urls:
            if jar.startswith("http://") or jar.startswith("https://"):
                resolved_jars.append(self._download_jar(jar))
            else:
                resolved_jars.append(jar)
        return resolved_jars

    @log_exceptions
    def _download_jar(self, url: str) -> str:
        """Downloads a JAR file from a URL."""
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme not in HTTP_SCHEMES:
            raise ValueError("Invalid URL scheme. Only HTTP and HTTPS are allowed.")
        os.makedirs(self.jar_dir, exist_ok=True)
        file_name = os.path.join(self.jar_dir, os.path.basename(parsed_url.path))
        try:
            response = requests.get(url, stream=True, timeout=REQUEST_TIMEOUT)
            if response.status_code == HTTP_STATUS_OK:
                with open(file_name, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        f.write(chunk)
                return file_name
            else:
                raise SparkAppsException(f"Failed to download file, HTTP status code: {response.status_code}")
        except Exception as e:
            log.error(f"Failed to download JAR from {url}: {e}")
            raise

    @log_exceptions
    def _build_maven_packages(self) -> str:
        """Constructs the Maven artifact string for Delta Lake."""
        delta_maven_artifact = DELTA_MAVEN_FORMAT.format(
            scala_version=self.scala_version, delta_version=self.delta_version
        )
        all_artifacts = [delta_maven_artifact, *self.extra_packages] if self.enable_delta_jar else self.extra_packages
        return ",".join(all_artifacts)



from pyspark.sql.types import StringType, StructField, StructType, TimestampType

from seedspark.sparkapp import SparkApps


class ExploreMusicDatasetApp(SparkApps):
    # def __init__(
    #     self,
    #     app_name: str = "ExploreMusicDatasetApp",
    #     extra_packages: Optional[List[str]] = None,
    #     extra_jars: Optional[List[str]] = None,
    #     extra_configs: Optional[Dict[str, Any]] = None,
    #     spark_master: Optional[str] = None,
    #     environment="staging",
    #     log_env: bool = True,
    #     existing_spark_session: Optional[SparkSession] = None,  # New parameter to accept an existing session
    # ):
    #     super().__init__(
    #         app_name,
    #         extra_packages=extra_packages,
    #         extra_jars=extra_jars,
    #         extra_configs=extra_configs,
    #         spark_master=spark_master,
    #         environment=environment,
    #         log_env=log_env,
    #         existing_spark_session=existing_spark_session,
    #     )

    def load_data(self):
        data_path = "/Users/chethanuk/Downloads/lastfm-dataset-1K/userid-timestamp-artid-artname-traid-traname.tsv"
        schema = StructType(
            [
                StructField("userid", StringType(), True),
                StructField("timestamp", TimestampType(), True),
                StructField("artist_id", StringType(), True),
                StructField("artist_name", StringType(), True),
                StructField("song_id", StringType(), True),
                StructField("song_name", StringType(), True),
            ]
        )

        data = self.spark.read.option("sep", "\t").option("header", "false").schema(schema).csv(data_path)
        print(f"Columns: {data.columns}")
        print(f"Schema: {data.schema}")
        return data

    def execute(self, profile=False):
        df = self.load_data()

        df.registerTempTable("songs_raw")

        self.spark.sql(
            """
                       SELECT userid, COUNT(1)
                       FROM songs_raw
                       GROUP BY userid
                       """
        ).show()

        df.show(100, truncate=False)
        if profile:
            from ydata_profiling import ProfileReport

            report = ProfileReport(
                df, title="Music dataset", infer_dtypes=False, interactions=None, missing_diagrams=None
            )
            report_html = report.to_html()
            with open("report.html", "w") as f:
                f.write(report_html)

            # Count the number of lines
            line_count = df.count()

            print(f"Number of lines in the file: {line_count}")
            return line_count


__main__ = ExploreMusicDatasetApp(app_name="ExploreMusicDatasetApp").execute()



from onetl.strategy import SnapshotStrategy

from seedspark.apps import SparkDeltaClickhouseApp


class WeekendMetrics:
    """
    Class to compute weekend metrics using SparkDeltaClickhouseApp.
    """

    def __init__(self, app_name="WeekendMetricsApp", environment="prod"):
        self.clickhouse_app = SparkDeltaClickhouseApp(app_name=app_name, environment=environment)

    def execute(self):
        """
        Execute specific metrics calculations for weekends and return DataFrame.
        """
        # Example SQL query
        query = "SELECT AirportID, Name, City, Country, Timezone FROM airports"
        print(self.clickhouse_app.clickhouse)
        return self.clickhouse_app.execute(query)


def run_snapshot_example():
    from onetl.db import DBReader

    clickhouse_app = SparkDeltaClickhouseApp(
        app_name="weekend_taxi_snapshot_read", environment="local", extra_packages=["org.xerial:sqlite-jdbc:3.45.1.0"]
    )
    table = f"{clickhouse_app.clickhouse_db}.airports"
    print(f"Reading table: {table}")
    reader = DBReader(
        connection=clickhouse_app.clickhouse,
        source=table,
        # SELECT AirportID, Name, City, Country, Timezone FROM airports
        columns=["AirportID", "Name", "City", "Country", "Timezone"],
        # NOTE: hwm cannot be used with SnapshotStrategy. Only with IncrementalStrategy
        # Also NOTE: Timezone is partition column
        # hwm=DBReader.AutoDetectHWM(name="clickhouse_hwm_name", expression="Timezone"),
    )

    with SnapshotStrategy():
        print(f"Fetching all rows with SnapshotStrategy for reader: {reader}")
        df = reader.run()
        df.show(2)

        # JDBC URL for SQLite
        jdbc_url = "jdbc:sqlite:airports_database.db"

        # Write to SQLite using JDBC
        df.write.format("jdbc").option("url", jdbc_url).option("dbtable", "airports").option(
            "driver", "org.sqlite.JDBC"
        ).mode("overwrite").save()

    # with IncrementalStrategy():
    #     df = reader.run()
    #     df.show(2)
    #     print(f"df.count(): {df.count()}")


if __name__ == "__main__":
    # Example 1: Run snapshot example
    print("Running snapshot example")
    run_snapshot_example()
    # Example 2: Run weekend metrics with SQL example
    print("Running weekend metrics example")
    weekendApp = WeekendMetrics(app_name="weekend_taxi_pipeline")
    weekendApp.execute()



from seedspark.configs.clickhouse import (
    CIClickHouseConfig,
    LocalClickHouseConfig,
    ProdClickHouseConfig,
    StagingClickHouseConfig,
)


class ConfigFactory:
    """A factory class for generating configurations based on environment."""

    def __init__(self, env_name: str) -> None:
        """Initialize ConfigFactory with environment name.

        Parameters
        ----------
        env_name (str): Name of the environment.

        """
        self._env_name = env_name.lower()
        self._clickhouse_config = None
        self._postgres_config = None

    @property
    def clickhouse(self):
        """Lazily initialize and return ClickHouse configuration.

        Returns
        -------
        ClickHouse configuration object.

        """
        if self._clickhouse_config is None:
            self._clickhouse_config = self._get_config(
                {
                    "local": LocalClickHouseConfig,
                    "dev": LocalClickHouseConfig,
                    "ci": CIClickHouseConfig,
                    "staging": StagingClickHouseConfig,
                    "prod": ProdClickHouseConfig,
                },
            )
        return self._clickhouse_config

    def _get_config(self, config_classes):
        """Get the configuration class based on environment.

        Parameters
        ----------
        config_classes (dict): Dictionary mapping environment names to configuration classes.

        Returns
        -------
        Configuration object based on environment.

        Raises
        ------
        ValueError: If the environment name is unknown.

        """
        config_class = config_classes.get(self._env_name)
        if not config_class:
            msg = f"Unknown environment: {self._env_name}"
            raise ValueError(msg)
        return config_class()



#
from functools import cached_property
from typing import Any, Dict, List, Optional

from loguru import logger as log
from pyspark.sql import SparkSession

from seedspark.configs import BaseClickHouseConfig, ConfigFactory
from seedspark.connections import ClickHouse
from seedspark.decorators import connection_check, log_exceptions, spark_session_management
from seedspark.sparkapp import SparkApps


class SparkDeltaClickhouseApp(SparkApps):
    def __init__(  # noqa: PLR0913
        self,
        app_name: str,
        clickhouse_config: BaseClickHouseConfig = None,
        extra_packages: Optional[List[str]] = None,
        extra_configs: Optional[Dict[str, Any]] = None,
        extra_jars: Optional[List[str]] = None,
        spark_master: Optional[str] = None,
        environment="staging",
        log_env: bool = True,
        existing_spark_session: Optional[SparkSession] = None,  # New parameter to accept an existing session
    ):
        self.extra_configs = {**(self._apply_delta_config()), **(extra_configs or {})}
        # self._apply_clickhouse_config(extra_configs)
        # Ensure extra_jars is a list of strings
        self.extra_jars = extra_jars or []
        self.environment = environment
        self._setup_jars()
        self.extra_packages = extra_packages or []
        self.extra_packages.extend(ClickHouse.get_packages())

        super().__init__(
            app_name,
            extra_packages=extra_packages,
            extra_jars=self.extra_jars,
            extra_configs=self.extra_configs,
            spark_master=spark_master,
            log_env=log_env,
            environment=self.environment,
            existing_spark_session=existing_spark_session,
        )
        print(self.spark.version)

        if clickhouse_config is None:
            self.clickhouse_configs = self.configs.clickhouse
        else:
            self.clickhouse_configs = clickhouse_config

    @log_exceptions
    def _setup_jars(self):
        # clickhouse_spark_jar_url = "your_clickhouse_spark_jar_url"
        clickhouse_jdbc_jar_url = (
            # "https://repo1.maven.org/maven2/com/clickhouse/clickhouse-jdbc/0.6.0/clickhouse-jdbc-0.6.0-all.jar"
            "https://github.com/ClickHouse/clickhouse-java/releases/download/v0.6.0-patch4/clickhouse-jdbc-0.6.0-patch4-all.jar"
        )
        self.extra_jars.extend([clickhouse_jdbc_jar_url])

    def _apply_clickhouse_config(self, clickhouse_configs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        _catalog_configs = {
            "spark.sql.catalog.clickhouse": "xenon.clickhouse.ClickHouseCatalog",
            "spark.sql.catalog.clickhouse.host": self.clickhouse_config.host,
            "spark.sql.catalog.clickhouse.protocol": self.clickhouse_config.protocol,
            "spark.sql.catalog.clickhouse.http_port": self.clickhouse_config.http_port,
            "spark.sql.catalog.clickhouse.user": self.clickhouse_config.user,
            "spark.sql.catalog.clickhouse.password": self.clickhouse_config.password,
            "spark.sql.catalog.clickhouse.database": self.clickhouse_config.database,
            "spark.sql.catalog.clickhouse.option.ssl": self.clickhouse_config.ssl,
        }
        return {**_catalog_configs, **(clickhouse_configs or {})}

    def _apply_delta_config(self, delta_configs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Applies Delta configurations to the provided dictionary.

        Args:
          delta_configs: A dictionary containing Delta configurations to be applied.

        Returns:
          A dictionary with combined catalog configurations and optional user-provided configurations.
        """
        _catalog_configs = {
            "spark.sql.extensions": "io.delta.sql.DeltaSparkSessionExtension",
            "spark.sql.catalog.spark_catalog": "org.apache.spark.sql.delta.catalog.DeltaCatalog",
        }
        return {**_catalog_configs, **(delta_configs or {})}

    @property  # Cache the ClickHouse connection
    def clickhouse(self) -> ClickHouse:
        """Establishes a connection to ClickHouse.

        Returns:
            ClickHouseConnection: A connection object.
        """
        return ClickHouse(
            spark=self.spark,
            host=self.clickhouse_config.host,
            port=self.clickhouse_config.http_port,
            user=self.clickhouse_config.user,
            password=self.clickhouse_config.password,
            database=self.clickhouse_config.database,
            extra={"ssl": str(self.clickhouse_config.ssl).lower()},
        )

    @cached_property
    def configs(self) -> ConfigFactory:  # Type hint can clarify
        return ConfigFactory(self.environment)  # Assumes environment remains consistent

    @cached_property
    def clickhouse_config(self) -> BaseClickHouseConfig:
        return self.configs.clickhouse

    @cached_property
    def clickhouse_db(self) -> BaseClickHouseConfig:
        return self.configs.clickhouse.database

    @cached_property
    def clickhouse_jdbc_url(self) -> str:
        """Computes the JDBC URL from the ClickHouse connection object.

        Returns:
            str: The JDBC URL.
        """
        return self.clickhouse.jdbc_url

    @cached_property
    @log_exceptions
    def clickhouse_reader(self, table, columns, partition_col):
        from onetl.db import DBReader

        return DBReader(
            connection=self.clickhouse,
            source=table,
            columns=columns,
            hwm=DBReader.AutoDetectHWM(name=f"clickhouse_{table}_hwm_name", expression=partition_col),
        )

    def pre_start(self):
        """
        Initialize configurations and check ClickHouse connection.
        """
        # self._init_clickhouse_client()
        if not self._check_connection():
            raise ConnectionError(f"Failed to connect to ClickHouse: {self.clickhouse}")

    def _check_connection(self) -> bool:
        """Check the ClickHouse database connection."""
        try:
            self.clickhouse.check()
            log.info(f"ClickHouse connection successful for url: {self.clickhouse_jdbc_url}")
            return True
        except Exception as e:
            log.error(f"ClickHouse connection check failed: {e}")
            return False

    @log_exceptions
    @spark_session_management
    @connection_check
    def execute(self, query: str) -> Any:
        log.info(f"Executing query in Clickhouse DB: {query}")
        df = self.clickhouse.sql(query)
        log.info(f"Schema of clickhouse dataframe: {df.schema}")
        df.show(5)
        return df

    @log_exceptions
    def post_stop(self):
        """Stop the Spark application and perform cleanup.
        NOTE: This is not working as expected
        Because Spark work asynchronusly, the cleanup is not working as expected. TODO: Fix this."""
        # super().stop_spark_session()
        # if self.clickhouse.check():
        #     self.clickhouse.close()
        log.info("Post-stop cleanup completed")



# table_name = "integers"

# psql_jdbc = "jdbc:postgresql://demoteddy-5340.6zw.aws-eu-west-1.cockroachlabs.cloud:26257/defaultdb?sslmode=verify-full"
# connection_properties = {
#     # Postgres
#     "driver": "org.postgresql.Driver",
#     "user": "demo",
#     "password": "fdVWHGGDfRzS2J3tjr2GCA",
#     "ssl": str(self.clickhouse_config.ssl)
#     # "sslmode": "strict",  # Can be 'strict', 'none', etc.,
# }

# postgres = Postgres(
#     host="demoteddy-5340.6zw.aws-eu-west-1.cockroachlabs.cloud",
#     user="demo",
#     port=26257,
#     password="fdVWHGGDfRzS2J3tjr2GCA",
#     database="defaultdb",
#     extra={"ssl": "true", "ssl" :"verify-full"},
#     spark=self.spark,
# )
# print(f"postgres: {postgres.jdbc_url}")
# reader = DBReader(
#     connection=postgres,
#     source="public.table_tennis_players_rank",
#     columns=["country", "player_name"],
#     # hwm=DBReader.AutoDetectHWM(name="psome_hwm_name", expression="id"),
# )
# with SnapshotStrategy():
#     df = reader.run()
#     df.show()
#     df.printSchema()

# # self.spark.read.jdbc(url=psql_jdbc, table="table_tennis_players_rank", properties=connection_properties).show()



import logging
import tempfile
import uuid
from pathlib import Path

import pytest
from pyspark import SparkConf
from pyspark.sql import SparkSession

logger = logging.getLogger("spark-test-logger")


class BaseSparkTest:
    """Base class for Spark tests with optimized Spark session initialization and teardown."""

    _spark = None

    @property
    def spark(self):
        """Lazily initialized Spark session. Creates a Spark session when first accessed."""
        if self._spark is None:
            logger.info("Initializing Spark session for testing")
            spark_conf = self.create_spark_conf()
            self._spark = SparkSession.builder.config(conf=spark_conf).getOrCreate()
            self.quiet_py4j(self._spark.sparkContext)
        return self._spark

    @classmethod
    def teardown_class(cls) -> None:
        """Class-level teardown to stop the Spark session."""
        if cls._spark:
            cls._spark.stop()
            cls._spark = None
            logger.info("Spark session for tests stopped")

    @staticmethod
    def create_spark_conf() -> SparkConf:
        """Creates a SparkConf object with settings optimized for testing."""
        return (
            SparkConf()
            .setAppName("spark-tests")
            .setMaster("local[2]")
            .set("spark.ui.showConsoleProgress", "false")
            .set("spark.sql.shuffle.partitions", "4")
        )

    @staticmethod
    def quiet_py4j(sc) -> None:
        """Reduces the log level of Spark's py4j to minimize noise during testing."""
        logging.getLogger("py4j").setLevel(logging.ERROR)
        sc.setLogLevel("ERROR")

    # @pytest.fixture(scope="function")
    @pytest.fixture(scope="class")
    def sparkSession(self):
        """Function-level fixture to manage temporary directories for each test."""
        temp_dir = tempfile.TemporaryDirectory()
        self.test_path = Path(temp_dir.name)
        yield self.spark
        temp_dir.cleanup()

    @staticmethod
    def random_uuid() -> str:
        """Generates a random UUID."""
        return str(uuid.uuid4())



import os
from dataclasses import dataclass


@dataclass
class BaseClickHouseConfig:
    """Base configuration for ClickHouse with default settings.
    Attributes are set using properties with fallbacks to default values.

    Attributes
    ----------
    HOST : str
        The host address of the ClickHouse server.
    PROTOCOL : str
        The protocol used to connect to ClickHouse (e.g., http).
    HTTP_PORT : str
        The port for HTTP connections to ClickHouse.
    USER : str
        Username for ClickHouse authentication.
    PASSWORD : str
        Password for ClickHouse authentication.
    DATABASE : str
        The default database to use in ClickHouse.

    """

    def get_conf(self, var_name: str, default: str) -> str:
        return os.getenv(var_name, default)

    @property
    def host(self) -> str:
        #
        return self.get_conf("CLICKHOUSE_HOST", "github.demo.altinity.cloud")

    @property
    def protocol(self) -> str:
        return self.get_conf("CLICKHOUSE_PROTOCOL", "https")

    @property
    def http_port(self) -> int:
        return int(self.get_conf("CLICKHOUSE_JDBC_PORT", "8443"))

    @property
    def user(self) -> str:
        return self.get_conf("CLICKHOUSE_USER", "demo")

    @property
    def password(self) -> str:
        return self.get_conf("CLICKHOUSE_PASSWORD", "demo")

    @property
    def database(self) -> str:
        return self.get_conf("CLICKHOUSE_DATABASE", "default")

    @property
    def ssl(self) -> bool:
        return True


class LocalClickHouseConfig(BaseClickHouseConfig):
    """Local environment configuration for ClickHouse."""


class CIClickHouseConfig(BaseClickHouseConfig):
    """CI environment configuration for ClickHouse."""


class StagingClickHouseConfig(BaseClickHouseConfig):
    """Staging environment configuration for ClickHouse."""


class ProdClickHouseConfig(BaseClickHouseConfig):
    """Production configuration for ClickHouse, overrides from environment variables."""

    @property
    def ssl(self) -> bool:
        return True



from typing import Optional

import sqlglot
from pyspark.sql.types import StringType, StructField, StructType, TimestampType
from sqlglot.errors import ErrorLevel

from seedspark.sparkapp import SparkApps


class LastFMSessionsApp(SparkApps):
    """
    Session Top 50 approach:
    ------------------------
    Session - One or More songs by UserID
        Song started 20min of prevSong start_time
        Time delta between two rows is not greather than 20mins


    Find Top 50 longest sessions by tracks count
        Top 50 longest sesions = Sessions with highest Count
    # Also Remove duplicates before doing track count top 50 filter

    List of Top 10 songs in Top 50

    """

    def load_data(self, data_path: Optional[str] = None):
        if not data_path:
            data_path = (
                "/Users/chethanuk//Downloads/lastfm-dataset-1K/userid-timestamp-artid-artname-traid-traname.tsv"
            )
            # TODO: Update or Replace with actual path of music_sessions_data.tsv
            # data_path = "./tests/fm_music_top_songs/it/it_music_sessions_data.tsv"
        schema = StructType(
            [
                StructField("userid", StringType(), True),
                StructField("timestamp", TimestampType(), True),
                StructField("artist_id", StringType(), True),
                StructField("artist_name", StringType(), True),
                StructField("song_id", StringType(), True),
                StructField("song_name", StringType(), True),
            ]
        )

        data = self.spark.read.option("sep", "\t").option("header", "false").schema(schema).csv(data_path)
        print(f"Columns: {data.columns}")
        print(f"Schema: {data.schema}")
        return data

    @staticmethod
    def generate_sessions_sql():
        return """
                WITH song_prev_data AS (
                                SELECT
                                    userid,
                                    timestamp,
                                    song_name,
                                    LAG(timestamp) OVER (PARTITION BY userid ORDER BY timestamp) as prev_timestamp
                                FROM
                                    songs_raw
                            ),
                    session_differences AS (
                        SELECT
                            userid,
                            timestamp,
                            song_name,
                            prev_timestamp,
                            CASE
                                WHEN prev_timestamp IS NULL OR
                                    (unix_timestamp(timestamp) - unix_timestamp(prev_timestamp)) / 60 > 20 THEN 1
                                ELSE 0
                            END as new_session_flag
                        FROM
                            song_prev_data
                    ),
                    sessions_with_marking AS (
                        SELECT
                            *,
                            SUM(new_session_flag) OVER (PARTITION BY userid ORDER BY timestamp) as session_id
                        FROM
                            session_differences
                    )
                """

    @staticmethod
    def top_50_sessions_sql():
        return """
            session_counts AS (
                SELECT
                    userid,
                    session_id,
                    COUNT(*) as song_count
                FROM
                    sessions_with_marking
                GROUP BY
                    userid, session_id
            ),
            top_sessions AS (
                SELECT
                    userid,
                    session_id,
                    song_count,
                    -- TODO: Based on requirment change Rank to DENSE_RANK
                    DENSE_RANK() OVER (PARTITION BY userid ORDER BY song_count DESC) as rank
                FROM
                    session_counts
                WHERE
                    song_count > 1 -- Verify: Filter sessions with at least more than one song
            ),
            filtered_sessions AS (
                SELECT
                    a.userid,
                    a.session_id,
                    -- a.artist_name,
                    a.song_name,
                    a.timestamp
                FROM
                    sessions_with_marking a
                JOIN
                    top_sessions b
                ON
                    a.userid = b.userid AND a.session_id = b.session_id
                WHERE
                    b.rank <= 50
            )
        """

    def parse_query(self, query: str):
        return sqlglot.transpile(query, read="spark", write="spark", error_level=ErrorLevel.RAISE)[0]

    def query(self, top_n: int = 10):
        sql_query = f"""
            {self.generate_sessions_sql()},\n
            {self.top_50_sessions_sql()}\n
            SELECT
                song_name,
                COUNT(song_name) as song_count
            FROM
                filtered_sessions
            GROUP BY song_name
            ORDER BY
                song_count DESC
            LIMIT {top_n}
            """

        return self.parse_query(sql_query)

    def execute(self, data_path: Optional[str] = None, output_path: Optional[str] = None, top_n: int = 10):
        df = self.load_data(data_path)
        df.createOrReplaceTempView("songs_raw")

        df_top_n_sessions = self.spark.sql(self.query(top_n))
        print(f"Query: {self.query()}")

        # df_top_n_sessions.explain(mode='simple')
        df_top_n_sessions.explain(mode="formatted")

        df_top_n_sessions.cache().show(100, truncate=False)

        # Write as JSON
        if output_path:
            # df_top_n_sessions.write.mode("overwrite").json(output_path)
            # For local testing - writing as single file
            # json.decoder.JSONDecodeError: Extra data: line 2 column 1
            df_top_n_sessions.coalesce(1).toPandas().to_json(output_path, orient="records")


if __name__ == "__main__":
    LastFMSessionsApp(app_name="ExploreMusicDatasetApp").execute(output_path="./datasets/top_n_songs.json", top_n=15)



from __future__ import annotations

import warnings
from typing import ClassVar

from onetl._util.classproperty import classproperty
from onetl.connection.db_connection.jdbc_connection import JDBCConnection
from onetl.connection.db_connection.jdbc_mixin.options import JDBCOptions
from onetl.hooks import slot, support_hooks
from onetl.impl import GenericOptions

from seedspark.connections.clickhouse_dialect import ClickhouseDialect


class ClickHouseExtra(GenericOptions):
    class Config:
        extra = "allow"
        ssl = "true"


@support_hooks
class ClickHouse(JDBCConnection):
    """ClickHouse JDBC connection. |support_hooks|

    Based on Maven package ``com.clickhouse:clickhouse-jdbc:0.6.0``
    (official ClickHouse JDBC driver).

    ... [Additional documentation and version compatibility details]

    Parameters
    ----------
    host : str
        Host of ClickHouse database.

    port : int, default: ``8123``
        HTTP port of ClickHouse database.

    user : str
        User with access to the database.

    password : str
        Password for database connection.

    database : str
        Target database in ClickHouse.

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session.

    extra : dict, default: ``None``
        Additional connection parameters.

    ... [Additional parameter documentation]

    Examples
    --------
    ... [Usage examples]
    """

    database: str
    port: int = 8123  # Default port for ClickHouse
    extra: ClickHouseExtra = ClickHouseExtra()

    Extra = ClickHouseExtra
    Dialect = ClickhouseDialect
    DRIVER: ClassVar[str] = "com.clickhouse.jdbc.ClickHouseDriver"

    @slot
    @classmethod
    def get_packages(cls) -> list[str]:
        """Get Maven packages for Spark."""
        return ["com.clickhouse:clickhouse-jdbc:0.6.0"]

    @classproperty
    def package(cls) -> str:
        """Get package name to be downloaded by Spark."""
        msg = "`Clickhouse.package` will be removed in 1.0.0, use `Clickhouse.get_packages()` instead"
        warnings.warn(msg, UserWarning, stacklevel=3)
        return "com.clickhouse:clickhouse-jdbc:0.6.0"

    @property
    def jdbc_url(self) -> str:
        extra = self.extra.dict(by_alias=True)
        parameters = "&".join(f"{k}={v}" for k, v in sorted(extra.items()))
        return f"jdbc:clickhouse://{self.host}:{self.port}/{self.database}?{parameters}".rstrip("?")

    @property
    def instance_url(self) -> str:
        return f"{super().instance_url}/{self.database}"

    def _options_to_connection_properties(self, options: JDBCOptions):
        return super()._options_to_connection_properties(options)
